# Copyright (c) 2013-2014 Will Thames <will@thames.id.au>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from ansiblelater.rule import RuleBase


class CheckGitHasVersion(RuleBase):
    rid = "ANS120"
    description = "Git checkouts should use explicit version"
    helptext = "git checkouts should point to an explicit commit or tag, not `latest`"
    types = ["playbook", "task", "handler"]

    def check(self, candidate, settings):
        tasks, errors = self.get_normalized_tasks(candidate, settings)

        if not errors:
            for task in tasks:
                if (
                    task["action"]["__ansible_module__"] == "git"
                    and task["action"].get("version", "HEAD") == "HEAD"
                ):
                    errors.append(self.Error(task["__line__"], self.helptext))

        return self.Result(candidate.path, errors)



# Original code written by the authors of ansible-lint

import functools

from ansiblelater.rule import RuleBase

SORTER_TASKS = (
    "name",
    # "__module__",
    # "action",
    # "args",
    None,  # <-- None include all modules that not using action and *
    # "when",
    # "notify",
    # "tags",
    "block",
    "rescue",
    "always",
)


class CheckKeyOrder(RuleBase):
    rid = "ANS129"
    description = "Check for recommended key order"
    helptext = "{type} key order can be improved to `{sorted_keys}`"
    types = ["playbook", "task", "handler"]

    def check(self, candidate, settings):
        errors = []
        tasks, err = self.get_normalized_tasks(candidate, settings)

        if err:
            return self.Result(candidate.path, err)

        for task in tasks:
            is_sorted, keys = self._sort_keys(task.get("__raw_task__"))
            if not is_sorted:
                errors.append(
                    self.Error(
                        task["__line__"],
                        self.helptext.format(type="task", sorted_keys=", ".join(keys)),
                    )
                )

        if candidate.kind == "playbook":
            tasks, err = self.get_tasks(candidate, settings)

            if err:
                return self.Result(candidate.path, err)

            for task in tasks:
                is_sorted, keys = self._sort_keys(task)
                if not is_sorted:
                    errors.append(
                        self.Error(
                            task["__line__"],
                            self.helptext.format(type="play", sorted_keys=", ".join(keys)),
                        )
                    )

        return self.Result(candidate.path, errors)

    @staticmethod
    def _sort_keys(task):
        if not task:
            return True, []

        keys = [str(key) for key in task if not key.startswith("_")]
        sorted_keys = sorted(keys, key=functools.cmp_to_key(_task_property_sorter))

        return (keys == sorted_keys), sorted_keys


def _task_property_sorter(property1, property2):
    """Sort task properties based on SORTER."""
    v_1 = _get_property_sort_index(property1)
    v_2 = _get_property_sort_index(property2)
    return (v_1 > v_2) - (v_1 < v_2)


def _get_property_sort_index(name):
    """Return the index of the property in the sorter."""
    a_index = -1
    for i, v in enumerate(SORTER_TASKS):
        if v == name:
            return i
        if v is None:
            a_index = i
    return a_index



"""Global utils collection."""

import contextlib
import re
import sys
from contextlib import suppress
from functools import lru_cache

import yaml
from ansible.plugins.loader import module_loader

from ansiblelater import logger

try:
    import ConfigParser as configparser  # noqa
except ImportError:
    import configparser  # noqa

LOG = logger.get_logger(__name__)


def count_spaces(c_string):
    leading_spaces = 0
    trailing_spaces = 0

    for _i, e in enumerate(c_string):
        if not e.isspace():
            break
        leading_spaces += 1

    for _i, e in reversed(list(enumerate(c_string))):
        if not e.isspace():
            break
        trailing_spaces += 1

    return (leading_spaces, trailing_spaces)


def lines_ranges(lines_spec):
    if not lines_spec:
        return None
    result = []
    for interval in lines_spec.split(","):
        (start, end) = interval.split("-")
        result.append(range(int(start), int(end) + 1))
    return result


def is_line_in_ranges(line, ranges):
    return not ranges or any(line in r for r in ranges)


def safe_load(string):
    """
    Parse the provided string returns a dict.

    :param string: A string to be parsed.
    :returns: dict

    """
    with suppress(yaml.scanner.ScannerError):
        return yaml.safe_load(string) or {}


@contextlib.contextmanager
def open_file(filename, mode="r"):
    """
    Open the provide file safely and returns a file type.

    :param filename: A string containing an absolute path to the file to open.
    :param mode: A string describing the way in which the file will be used.
    :returns: file type

    """
    with open(filename, mode) as stream:
        yield stream


def add_dict_branch(tree, vector, value):
    key = vector[0]
    tree[key] = (
        value if len(vector) == 1 else add_dict_branch(tree.get(key, {}), vector[1:], value)
    )
    return tree


def has_jinja(value):
    """Return true if a string seems to contain jinja templating."""
    re_has_jinja = re.compile(r"{[{%#].*[%#}]}", re.DOTALL)
    return bool(isinstance(value, str) and re_has_jinja.search(value))


def has_glob(value):
    """Return true if a string looks like having a glob pattern."""
    re_has_glob = re.compile("[][*?]")
    return bool(isinstance(value, str) and re_has_glob.search(value))


def sysexit(code=1):
    sys.exit(code)


def sysexit_with_message(msg, code=1):
    LOG.critical(msg)
    sysexit(code)


class Singleton(type):
    """Meta singleton class."""

    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]


@lru_cache
def load_plugin(name):
    """Return loaded ansible plugin/module."""
    loaded_module = module_loader.find_plugin_with_context(
        name,
        ignore_deprecated=True,
        check_aliases=True,
    )
    if not loaded_module.resolved and name.startswith("ansible.builtin."):
        # fallback to core behavior of using legacy
        loaded_module = module_loader.find_plugin_with_context(
            name.replace("ansible.builtin.", "ansible.legacy."),
            ignore_deprecated=True,
            check_aliases=True,
        )
    return loaded_module



from ansiblelater.rule import RuleBase


class CheckInstallUseLatest(RuleBase):
    rid = "ANS109"
    description = "Package installs should use present, not latest"
    helptext = "package installs should use `state=present` with or without a version"
    types = ["playbook", "task", "handler"]

    def check(self, candidate, settings):
        tasks, errors = self.get_normalized_tasks(candidate, settings)
        package_managers = [
            "yum",
            "apt",
            "dnf",
            "homebrew",
            "pacman",
            "openbsd_package",
            "pkg5",
            "portage",
            "pkgutil",
            "slackpkg",
            "swdepot",
            "zypper",
            "bundler",
            "pip",
            "pear",
            "npm",
            "yarn",
            "gem",
            "easy_install",
            "bower",
            "package",
            "apk",
            "openbsd_pkg",
            "pkgng",
            "sorcery",
            "xbps",
        ]

        if not errors:
            for task in tasks:
                if (
                    task["action"]["__ansible_module__"] in package_managers
                    and task["action"].get("state") == "latest"
                ):
                    errors.append(self.Error(task["__line__"], self.helptext))

        return self.Result(candidate.path, errors)
